@@ -27,11 +27,13 @@
 
 namespace wvWare
 {
-    namespace Word97 { class PHE; }
+    namespace Word97 { class PHE; class BTE; }
     namespace Word95
     {
         class PHE;
-        Word97::PHE toWord97( const Word95::PHE& phe );  // fake, to make gcc 3.4 happy :-(
+        class BTE;
+        Word97::PHE toWord97( const Word95::PHE& phe ); // fake, to make gcc 3.4 happy :-(
+        Word97::BTE toWord97( const Word95::BTE& s );   // fake, to make gcc 3.4 happy :-(
     }
 
     class OLEStreamReader;
@@ -187,6 +189,7 @@ namespace wvWare
     public:
         PLCF( U32 length, OLEStreamReader *reader, bool preservePos = false );
         PLCF( U32 length, const U8* ptr );
+        PLCF( const PLCF<T>& rhs );
         ~PLCF();
 
         PLCFIterator<T> at( unsigned int index ) const;
@@ -199,8 +202,7 @@ namespace wvWare
 
         void dumpCPs() const;
     private:
-        // don't copy or assign it
-        PLCF( const PLCF<T>& rhs );
+        // don't assign it
         PLCF<T>& operator=( const PLCF<T>& rhs );
 
         // An empty default constructor for the convertPLCF friend. Don't use it
@@ -237,6 +239,15 @@ namespace wvWare
             m_items.push_back( new T( ptr ) );
     }
 
+    template<class T>
+    PLCF<T>::PLCF( const PLCF<T>& rhs ) : m_indices( rhs.m_indices )
+    {
+        typename std::vector<T*>::const_iterator it = rhs.m_items.begin();
+        typename std::vector<T*>::const_iterator end = rhs.m_items.end();
+        for ( ; it != end; ++it )
+            m_items.push_back( new T( **it ) );
+    }
+
     template<class T>
     PLCF<T>::~PLCF()
     {
@@ -394,7 +405,7 @@ namespace wvWare
     }
 
 
-    template<class PHE> class BX;
+    template<class PHE> struct BX;
     template<class Offset> class FKP;
     template<class Offset> class FKPIterator;
     FKP< BX<Word97::PHE> >* convertFKP( const FKP< BX<Word95::PHE> >& old );
@@ -406,14 +417,14 @@ namespace wvWare
     public:
         FKP( OLEStreamReader* reader, bool preservePos = false );
         FKP( const U8* ptr );
+        FKP( const FKP<Offset>& rhs );
         ~FKP() { delete [] m_rgfc; delete [] m_rgb; delete [] m_fkp; }
 
         unsigned int crun() const { return m_crun; }
         bool isEmpty() const { return m_crun == 0; }
 
     private:
-        // don't copy or assign it
-        FKP( const FKP<Offset>& rhs );
+        // don't assign it
         FKP<Offset>& operator=( const FKP<Offset>& rhs );
 
         // An empty default constructor for the convertFKP friend. Don't use it
@@ -449,7 +460,7 @@ namespace wvWare
 
         // store the rest of the FKP in an internal array for later use
         const U16 length = 511 - m_internalOffset;
-        m_fkp = new U8 [ length ];  // 511, because we don't need crun
+        m_fkp = new U8[ length ];  // 511, because we don't need crun
         for ( U16 i = 0; i < length; ++i )
             m_fkp[ i ] = reader->readU8();
 
@@ -474,11 +485,27 @@ namespace wvWare
 
         // store the rest of the FKP in an internal array for later use
         U16 length = 511 - m_internalOffset;
-        m_fkp = new U8 [ length ];  // 511, because we don't need crun
+        m_fkp = new U8[ length ];  // 511, because we don't need crun
         for ( U16 i = 0; i < length; ++i, ++ptr )
             m_fkp[ i ] = *ptr;
     }
 
+    template<class Offset>
+    FKP<Offset>::FKP( const FKP<Offset>& rhs ) :
+        m_crun( rhs.m_crun ), m_internalOffset( rhs.m_internalOffset )
+    {
+        m_rgfc = new U32[ m_crun + 1 ];
+        ::memcpy( m_rgfc, rhs.m_rgfc, sizeof( U32 ) * ( m_crun + 1 ) );
+
+        m_rgb = new Offset[ m_crun ];
+        for ( U8 i = 0; i < m_crun; ++i )
+            m_rgb[ i ] = rhs.m_rgb[ i ];
+
+        const U16 length = 511 - m_internalOffset;
+        m_fkp = new U8[ length ];  // 511, because we don't need crun
+        ::memcpy( m_fkp, rhs.m_fkp, sizeof( U8 ) * length );
+    }
+
 
     // Attention: This iterator has a non standard behavior of the current()
     // method. Use it that way: for( ; !it.atEnd(); ++it)
@@ -525,8 +552,20 @@ namespace wvWare
             const U8 tmp = m_fkp.m_rgb[ m_index ].offset;
             // Now we have to calculate the real offset and then locate it
             // within our cached array...
-            if ( tmp != 0 )
-                return &m_fkp.m_fkp[ tmp * 2 - m_fkp.m_internalOffset ];
+            if ( tmp != 0 ) {
+                const int pos = tmp * 2 - m_fkp.m_internalOffset;
+                if ( pos < 0 ) {
+                    wvlog << "ERROR: FKP internalOffset (" << m_fkp.m_internalOffset << ") is bigger than " <<
+                        "2*" << (int)tmp << ", FKP array index would be negative!" << std::endl;
+                    return 0;
+                } else if ( pos >= 511 - m_fkp.m_internalOffset ) {
+                    wvlog << "ERROR: FKP array index (" << pos << " is bigger than allocated size ("
+                          << 511 - m_fkp.m_internalOffset << ")" << std::endl;
+                    return 0;
+                } else {
+                    return &m_fkp.m_fkp[ pos ];
+                }
+            }
         }
         return 0;
     }
